@@ -113,7 +113,8 @@ sub getFirstChildElement {
     my $n = $node->firstChild;
     while ($n && $n->nodeType != XML_ELEMENT_NODE) {
       $n = $n->nextSibling; }
-    return $n; } }
+    return $n; }
+  return; }
 
 # Find the nodes according to the given $xpath expression,
 # the xpath is relative to $node (if given), otherwise to the document node.
@@ -332,6 +333,7 @@ sub doctest_children {
 # It removes the `helper' attributes that store fonts, source box, etc.
 sub finalize {
   my ($self) = @_;
+  $self->pruneXMDuals;
   if (my $root = $self->getDocument->documentElement) {
     local $LaTeXML::FONT = $self->getNodeFont($root);
     $self->finalize_rec($root);
@@ -345,18 +347,23 @@ sub finalize_rec {
   my $declared_font       = $LaTeXML::FONT;
   my $desired_font        = $LaTeXML::FONT;
   my %pending_declaration = ();
+  if (my $comment = $node->getAttribute('_pre_comment')) {
+    $node->parentNode->insertBefore(XML::LibXML::Comment->new($comment), $node); }
+  if (my $comment = $node->getAttribute('_comment')) {
+    $node->parentNode->insertAfter(XML::LibXML::Comment->new($comment), $node); }
+
   if (my $font_attr = $node->getAttribute('_font')) {
     $desired_font        = $$self{node_fonts}{$font_attr};
-    %pending_declaration = $desired_font->relativeTo($LaTeXML::FONT);
-    if ($model->canHaveAttribute($qname, 'font')
-      && ($node->hasChildNodes || $node->getAttribute('_force_font'))
+    %pending_declaration = $desired_font->relativeTo($declared_font);
+    if (($node->hasChildNodes || $node->getAttribute('_force_font'))
       && scalar(keys %pending_declaration)) {
       foreach my $attr (keys %pending_declaration) {
-        $self->setAttribute($node, $attr => $pending_declaration{$attr})
-          if $model->canHaveAttribute($qname, $attr); }
-      $declared_font       = $desired_font;
-      %pending_declaration = (); } }
-
+        if ($model->canHaveAttribute($qname, $attr)) {
+          $self->setAttribute($node, $attr => $pending_declaration{$attr}{value});
+          # Merge to set the font currently in effect
+          $declared_font = $declared_font->merge(%{ $pending_declaration{$attr}{properties} });
+          delete $pending_declaration{$attr}; } }
+    } }
   local $LaTeXML::FONT = $declared_font;
   foreach my $child ($node->childNodes) {
     my $type = $child->nodeType;
@@ -379,7 +386,7 @@ sub finalize_rec {
         # Too late to do wrapNodes?
         my $text = $self->wrapNodes($FONT_ELEMENT_NAME, $child);
         foreach my $attr (keys %pending_declaration) {
-          $self->setAttribute($text, $attr => $pending_declaration{$attr}); }
+          $self->setAttribute($text, $attr => $pending_declaration{$attr}{value}); }
         $self->finalize_rec($text);    # Now have to clean up the new node!
       }
     } }
@@ -676,7 +683,7 @@ sub maybeCloseElement {
 
 # This closes all nodes until $node becomes the current point.
 sub closeToNode {
-  my ($self, $node) = @_;
+  my ($self, $node, $ifopen) = @_;
   my $model = $$self{model};
   my ($t, @cant_close) = ();
   my $n = $$self{node};
@@ -689,7 +696,8 @@ sub closeToNode {
   if ($t == XML_DOCUMENT_NODE) {    # Didn't find $node at all!!
     Error('malformed', $model->getNodeQName($node), $self,
       "Attempt to close " . Stringify($node) . ", which isn't open",
-      "Currently in " . $self->getInsertionContext); }
+      "Currently in " . $self->getInsertionContext) unless $ifopen;
+    return; }
   else {                            # Found node.
     Error('malformed', $model->getNodeQName($node), $self,
       "Closing " . Stringify($node) . " whose open descendents do not auto-close",
@@ -1019,9 +1027,9 @@ sub autoCollapseChildren {
 # insertion point (eg $$self{node}), but rather relative to nodes specified
 # in the arguments.
 
-# Set an attribute on a node, decoding the prefix, if any.
+# Set any allowed attribute on a node, decoding the prefix, if any.
 # Also records, and checks, any id attributes.
-# We _could_ check whether attribute is even allowed here? NOT YET.
+# [xml:id and namespaced attributes are always allowed]
 sub setAttribute {
   my ($self, $node, $key, $value) = @_;
   $value = $value->toAttribute if ref $value;
@@ -1030,8 +1038,13 @@ sub setAttribute {
       $value = $self->recordID($value, $node);    # Do id book keeping
       $node->setAttributeNS($LaTeXML::Common::XML::XML_NS, 'id', $value); }    # and bypass all ns stuff
     elsif ($key !~ /:/) {    # No colon; no namespace (the common case!)
-      $node->setAttribute($key => $value); }
-    else {
+                             # Ignore attributes not allowed by the model,
+                             # but accept "internal" attributes.
+      my $model = $$self{model};
+      my $qname = $model->getNodeQName($node);
+      if ($model->canHaveAttribute($qname, $key) || $key =~ /^_/) {
+        $node->setAttribute($key => $value); } }
+    else {                   # Accept any namespaced attributes
       my ($ns, $name) = $$self{model}->decodeQName($key);
       if ($ns) {             # If namespaced attribute (must have prefix!
         my $prefix = $node->lookupNamespacePrefix($ns);    # namespace already declared?
@@ -1099,6 +1112,82 @@ sub lookupID {
   my ($self, $id) = @_;
   return $$self{idstore}{$id}; }
 
+#======================================================================
+# Odd bit:
+# In an XMDual, in each branch (content, presentation) there will be atoms
+# that correspond to the input (one will be real, the other an XMRef to the first).
+# But also there will be additional "decoration" (delimiters, punctuation, etc on the presentation
+# side; other symbols, bindings, whatever, on the content side).
+# These decorations should NOT be subject to rewrite rules,
+# and in cross-linked parallel markup, they should be attributed to the
+# upper containing object's ID, rather than left dangling.
+#
+# To determine this, we mark all math nodes as to whether they are "visible" from
+# presentation, content or both (the default top-level being both).
+# Decorations are the nodes that are visible to only one mode.
+# Note that nodes that are not visible at all CAN occur (& do currently when the parser
+# creates XMDuals), pruneXMDuals (below) gets rid of them.
+
+# NOTE: This should ultimately be in a base Document class,
+# since it is also needed before conversion to parallel markup!
+sub markXMNodeVisibility {
+  my ($self) = @_;
+  my @xmath = $self->findnodes('//ltx:XMath/*');
+  foreach my $math (@xmath) {
+    foreach my $node ($self->findnodes('descendant-or-self::*[@_pvis or @_cvis]', $math)) {
+      $node->removeAttribute('_pvis');
+      $node->removeAttribute('_cvis'); } }
+  foreach my $math (@xmath) {
+    $self->markXMNodeVisibility_aux($math, 1, 1); }
+  return; }
+
+sub markXMNodeVisibility_aux {
+  my ($self, $node, $cvis, $pvis) = @_;
+  my $qname = $self->getNodeQName($node);
+  return if (!$cvis || $node->getAttribute('_cvis')) && (!$pvis || $node->getAttribute('_pvis'));
+  # Special case: for XMArg used to wrap "formal" arguments on the content side,
+  # mark them as visible as presentation as well.
+  $pvis = 1 if $cvis && ($qname eq 'ltx:XMArg');
+  $node->setAttribute('_cvis' => 1) if $cvis;
+  $node->setAttribute('_pvis' => 1) if $pvis;
+  if ($qname eq 'ltx:XMDual') {
+    my ($c, $p) = element_nodes($node);
+    $self->markXMNodeVisibility_aux($c, 1, 0) if $cvis;
+    $self->markXMNodeVisibility_aux($p, 0, 1) if $pvis; }
+  elsif ($qname eq 'ltx:XMRef') {
+    #    $self->markXMNodeVisibility_aux($self->realizeXMNode($node),$cvis,$pvis); }
+    my $id = $node->getAttribute('idref');
+    if (!$id) {
+      Warn('expected', 'id', $self, "Missing id on ltx:XMRef");
+      return; }
+    my $reffed = $self->lookupID($id);
+    if (!$reffed) {
+      Warn('expected', 'node', $self, "No node found with id=$id (referred to from ltx:XMRef)");
+      return; }
+    $self->markXMNodeVisibility_aux($reffed, $cvis, $pvis); }
+  else {
+    foreach my $child (element_nodes($node)) {
+      $self->markXMNodeVisibility_aux($child, $cvis, $pvis); } }
+  return; }
+
+# Reduce any ltx:XMDual's to just the visible branch, if the other is not visible
+# (according to markXMNodeVisibility)
+# If we could be 100% sure that the marking had stayed consistent (after various doc surgery)
+# we could avoid re-marking, but we'd better be sure before removing nodes!
+sub pruneXMDuals {
+  my ($self) = @_;
+  # RE-mark visibility!
+  $self->markXMNodeVisibility;
+  # will reversing keep from problems removing nodes from trees that already have been removed?
+  foreach my $dual (reverse $self->findnodes('descendant-or-self::ltx:XMDual')) {
+    my ($content, $presentation) = element_nodes($dual);
+    if (!$self->findnode('descendant-or-self::*[@_pvis or @_cvis]', $content)) {    # content never seen
+      $self->replaceTree($presentation, $dual); }
+    elsif (!$self->findnode('descendant-or-self::*[@_pvis or @_cvis]', $presentation)) {    # pres.
+      $self->replaceTree($content, $dual); }
+  }
+  return; }
+
 #**********************************************************************
 # Record the Box that created this node.
 sub setNodeBox {
